@@ -32,7 +32,6 @@ from builtins import *  # noqa
 import collections
 from datetime import datetime
 import functools
-from ipaddress import ip_address
 import logging
 import os
 import sys
@@ -82,14 +81,19 @@ def private(f):
     @functools.wraps(f)
     def wrapper(self, request):
         config = talisker.get_config()
-        if not request.access_route:
-            # this means something probably bugged in werkzeug, but let's fail
-            # gracefully
-            return Response('no client ip provided', status='403 Forbidden')
+        # talisker middleware provides this
+        ip_str = request.environ.get('CLIENT_ADDR')
+        if ip_str is None:
+            # fallback to werkzeug's handling
+            ip_str = force_unicode(request.access_route[-1])
 
-        ip_str = force_unicode(request.access_route[-1])
-        ip = ip_address(ip_str)
-        if ip.is_loopback or any(ip in network for network in config.networks):
+        if ip_str is None:
+            return Response(
+                'no client ip provided',
+                status='403 Forbidden',
+            )
+
+        if config.is_trusted_addr(ip_str):
             return f(self, request)
         else:
             msg = PRIVATE_BODY_RESPONSE_TEMPLATE.format(
@@ -97,6 +101,7 @@ def private(f):
                 force_unicode(request.remote_addr),
                 request.headers.get('x-forwarded-for'))
             return Response(msg, status='403 Forbidden')
+
     return wrapper
 
 
@@ -153,15 +158,9 @@ class StandardEndpointMiddleware(object):
     def __call__(self, environ, start_response):
         request = Request(environ)
         response = None
+
         if request.path.startswith(self.prefix):
-            path = request.path[len(self.prefix):].rstrip('/')
-            try:
-                funcname = self.urlmap.get(path, None)
-                func = getattr(self, funcname)
-            except (KeyError, AttributeError, TypeError):
-                pass
-            else:
-                response = func(request)
+            response = self.process_request(request, start_response)
 
         if response is None:
             # pass thru to the app
@@ -169,6 +168,23 @@ class StandardEndpointMiddleware(object):
         else:
             return response(environ, start_response)
 
+    def process_request(self, request, start_response):
+        config = talisker.get_config()
+
+        if config.status_interface:
+            iface, port = request.environ['gunicorn.socket'].getsockname()
+            if iface != config.status_interface:
+                return None
+
+        path = request.path[len(self.prefix):].rstrip('/')
+        try:
+            funcname = self.urlmap.get(path, None)
+            func = getattr(self, funcname)
+        except (KeyError, AttributeError, TypeError):
+            return None
+        else:
+            return func(request)
+
     def index(self, request):
         methods = []
         base = request.host_url.rstrip('/')
