@@ -32,8 +32,8 @@ from builtins import *  # noqa
 import collections
 import logging
 import time
-import shlex
 
+import psycopg2
 from psycopg2.extensions import cursor, connection
 
 try:
@@ -44,7 +44,6 @@ except ImportError:
 
 import talisker
 from talisker.util import get_rounded_ms
-from talisker.context import track_request_metric
 import talisker.sentry
 
 __all__ = [
@@ -69,23 +68,25 @@ def prettify_sql(sql):
         indent_tabs=False)
 
 
-def get_safe_connection_string(conn):
-    try:
-        try:
-            # 2.7+
-            params = conn.get_dsn_parameters()
-        except AttributeError:
-            params = dict(i.split('=') for i in shlex.split(conn.dsn))
-
-        params.setdefault('host', 'localhost')
-        return '{user}@{host}:{port}/{dbname}'.format(**params)
-    except Exception:
-        return 'could not parse dsn'
-
-
 class TaliskerConnection(connection):
     _logger = None
     _threshold = None
+    _explain = None
+    _safe_dsn = None
+    _safe_dsn_format = '{user}@{host}:{port}/{dbname}'
+
+    @property
+    def safe_dsn(self):
+        if self._safe_dsn is None:
+            try:
+                params = self.get_dsn_parameters()
+                params.setdefault('host', 'localhost')
+                self._safe_dsn = self._safe_dsn_format.format(**params)
+            except Exception:
+                self.logger.exception('Failed to parse DSN')
+                self._safe_dsn = 'could not parse dsn'
+
+        return self._safe_dsn
 
     @property
     def logger(self):
@@ -99,31 +100,52 @@ class TaliskerConnection(connection):
             self._threshold = talisker.get_config().slowquery_threshold
         return self._threshold
 
+    @property
+    def explain_breadcrumbs(self):
+        if self._explain is None:
+            self._explain = talisker.get_config().explain_sql
+        return self._explain
+
     def cursor(self, *args, **kwargs):
         kwargs.setdefault('cursor_factory', TaliskerCursor)
         return super().cursor(*args, **kwargs)
 
-    def _format_query(self, query):
+    def _format_query(self, query, vars):
         if callable(query):
             query = query()
         query = prettify_sql(query)
-        query = FILTERED if query is None else query
+        if query is None or vars is None:
+            return FILTERED
         return query
 
-    def _record(self, msg, query, duration):
-        track_request_metric('sql', duration)
+    def _record(self, msg, query, vars, duration, extra={}):
+        talisker.Context.track('sql', duration)
+
+        qdata = collections.OrderedDict()
+        qdata['duration_ms'] = duration
+        qdata['connection'] = self.safe_dsn
+        qdata.update(extra)
+
+        # grab a reference here, where super() works
+        base_connection = super()
 
         if self.query_threshold >= 0 and duration > self.query_threshold:
-            extra = collections.OrderedDict()
-            extra['trailer'] = self._format_query(query)
-            extra['duration_ms'] = duration
-            extra['connection'] = get_safe_connection_string(self)
-            self.logger.info('slow ' + msg, extra=extra)
+            formatted = self._format_query(query, vars)
+            self.logger.info(
+                'slow ' + msg, extra=dict(qdata, trailer=formatted))
 
         def processor(data):
-            data['data']['query'] = self._format_query(query)
-            data['data']['duration'] = duration
-            data['data']['connection'] = get_safe_connection_string(self)
+            qdata['query'] = self._format_query(query, vars)
+            if self.explain_breadcrumbs or talisker.Context.debug:
+                try:
+                    cursor = base_connection.cursor()
+                    cursor.execute('EXPLAIN ' + query, vars)
+                    plan = '\n'.join(l[0] for l in cursor.fetchall())
+                    qdata['plan'] = plan
+                except Exception as e:
+                    qdata['plan'] = 'could not explain query: ' + str(e)
+
+            data['data'].update(qdata)
 
         breadcrumb = dict(
             message=msg, category='sql', data={}, processor=processor)
@@ -133,22 +155,56 @@ class TaliskerConnection(connection):
 
 class TaliskerCursor(cursor):
 
+    def apply_timeout(self):
+        ctx_timeout = talisker.Context.deadline_timeout()
+        if ctx_timeout is None:
+            return None
+
+        ms = int(ctx_timeout * 1000)
+        super().execute(
+            'SET LOCAL statement_timeout TO %s', (ms,)
+        )
+        return ms
+
     def execute(self, query, vars=None):
+        extra = collections.OrderedDict()
+        timeout = self.apply_timeout()
+        if timeout is not None:
+            extra['timeout'] = timeout
         timestamp = time.time()
         try:
-            return super(TaliskerCursor, self).execute(query, vars)
+            return super().execute(query, vars)
+        except psycopg2.OperationalError as exc:
+            extra['pgcode'] = exc.pgcode
+            extra['pgerror'] = exc.pgerror
+            if exc.pgcode == '57014':
+                extra['timedout'] = True
+            raise
         finally:
             duration = get_rounded_ms(timestamp)
-            if vars is None:
-                query = None
-            self.connection._record('query', query, duration)
+            self.connection._record('query', query, vars, duration, extra)
 
     def callproc(self, procname, vars=None):
+        extra = collections.OrderedDict()
+        timeout = self.apply_timeout()
+        if timeout is not None:
+            extra['timeout'] = timeout
         timestamp = time.time()
         try:
-            return super(TaliskerCursor, self).callproc(procname, vars)
+            return super().callproc(procname, vars)
+        except psycopg2.OperationalError as exc:
+            extra['pgcode'] = exc.pgcode
+            extra['pgerror'] = exc.pgerror
+            if exc.pgcode == '57014':
+                extra['timedout'] = True
+            raise
         finally:
             duration = get_rounded_ms(timestamp)
             # no query parameters, cannot safely record
             self.connection._record(
-                'stored proc: {}'.format(procname), None, duration)
+                'stored proc: {}'.format(procname),
+                None,
+                vars,
+                duration,
+                extra,
+            )
