@@ -30,11 +30,11 @@ from __future__ import absolute_import
 from builtins import *  # noqa
 
 import collections
+from datetime import datetime
 import functools
 import itertools
 import logging
 import random
-import threading
 import warnings
 import time
 from future.moves.urllib.parse import (
@@ -50,17 +50,15 @@ from requests.adapters import HTTPAdapter
 import requests.exceptions
 import urllib3.exceptions
 from urllib3.util import Retry
-import werkzeug.local
 
 import talisker
-from talisker.context import track_request_metric
-from talisker import request_id
+from talisker import Context
 import talisker.metrics
 from talisker.util import (
-    context_local,
     get_errno_fields,
     module_dict,
     parse_url,
+    Local,
 )
 
 __all__ = [
@@ -71,17 +69,16 @@ __all__ = [
 ]
 
 
-STORAGE = threading.local()
+STORAGE = Local()
 STORAGE.sessions = {}
 HOSTS = module_dict()
-_local = context_local()
+DEBUG_HEADER = future.utils.text_to_native_str('X-Debug')
 
 
 def clear():
     if hasattr(STORAGE, 'sessions'):
         STORAGE.sessions.clear()
     HOSTS.clear()
-    werkzeug.local.release_local(_local)
 
 
 # storage for metric url state, as requests design allows for no other way
@@ -166,9 +163,15 @@ def send_wrapper(func):
 
     @functools.wraps(func)
     def send(request, **kwargs):
-        rid = request_id.get()
+        rid = Context.request_id
         if rid and config.id_header not in request.headers:
             request.headers[config.id_header] = rid
+        if Context.current.deadline:
+            deadline = datetime.utcfromtimestamp(Context.current.deadline)
+            formatted = deadline.isoformat() + 'Z'
+            request.headers[config.deadline_header] = formatted
+        if Context.debug:
+            request.headers[DEBUG_HEADER] = '1'
         try:
             return func(request, **kwargs)
         except Exception as e:
@@ -183,13 +186,19 @@ def request_wrapper(func):
     """Adds support for metric_name kwarg to session."""
     @functools.wraps(func)
     def request(method, url, **kwargs):
+        ctx = Context.current
         try:
-            _local.metric_api_name = kwargs.pop('metric_api_name', None)
-            _local.metric_host_name = kwargs.pop('metric_host_name', None)
+            ctx.metric_api_name = kwargs.pop('metric_api_name', None)
+            ctx.metric_host_name = kwargs.pop('metric_host_name', None)
             return func(method, url, **kwargs)
         finally:
-            del _local.metric_api_name
-            del _local.metric_host_name
+            # some requests errors can cause the context to be lost, and we
+            # should never fail due to this.
+            try:
+                del ctx.metric_api_name
+                del ctx.metric_host_name
+            except Exception:
+                pass
     request._request_wrapper = True
     return request
 
@@ -267,7 +276,7 @@ def metrics_response_hook(response, **kwargs):
 def record_request(request, response=None, exc=None):
     metadata = collect_metadata(request, response)
     if response:
-        track_request_metric('http', metadata['duration_ms'])
+        Context.track('http', metadata['duration_ms'])
 
     if exc:
         metadata.update(get_errno_fields(exc))
@@ -283,8 +292,9 @@ def record_request(request, response=None, exc=None):
         'view': metadata.get('view', 'unknown'),
     }
 
-    metric_api_name = getattr(_local, 'metric_api_name', None)
-    metric_host_name = getattr(_local, 'metric_host_name', None)
+    ctx = Context.current
+    metric_api_name = getattr(ctx, 'metric_api_name', None)
+    metric_host_name = getattr(ctx, 'metric_host_name', None)
     if metric_api_name is not None:
         labels['view'] = metric_api_name
     if metric_host_name is not None:
@@ -422,6 +432,12 @@ class TaliskerAdapter(HTTPAdapter):
                         'or two float/ints and a urllib3.Retry'
                     )
 
+        # enforce any context deadline
+        ctx_timeout = Context.deadline_timeout()
+        if ctx_timeout is not None:
+            connect = min(connect, ctx_timeout)
+            read = min(read, ctx_timeout)
+
         # ensure urllib3 timeout
         kwargs['timeout'] = (connect, read)
 
@@ -445,9 +461,9 @@ class TaliskerAdapter(HTTPAdapter):
 
             error = exc.args[0]
             if isinstance(error, urllib3.exceptions.MaxRetryError):
+                # we need the original urllib3 exception base error
                 error = error.reason
             try:
-                # we need the original urllib3 exception base error
                 request._retry = request._retry.increment(
                     request.method,
                     request.url,
@@ -473,7 +489,7 @@ class TaliskerAdapter(HTTPAdapter):
                 # which we do not want. So we explicitly reraise the original
                 # ConnectionError. In py3, this would not be desirable, as the
                 # exception context would be confused, by py2 does not do
-                # chained exceptins, so, erm, yay?
+                # chained exceptions, so, erm, yay?
                 if future.utils.PY3:
                     raise  # raises the original ConnectionError
                 else:
